import requests

BASE = 'http://127.0.0.1:5000/' # The endpoint of the API

''' Functions '''
def register():
	''' Give a username and password '''
	username = input('Please provide a username. ')
	if username == 'back': return False # To go back to previous layer

	display_name = input('Please provide a display name. ')
	if display_name == 'back': return False # To go back to previous layer

	while True:

		password1 = input('Please provide a password. ')
		if password1 == 'back': return False # To go back to previous layer

		password2 = input('Please repeat the password. ')
		if password2 == 'back': return False # To go back to previous layer

		if password1 == password2:
			password = password1
			break
		else:
			print('Sorry, the passwords do not match, please try again.')

	''' Adding friends '''
	print('If applicable, please provide usernames of up to five friends who already use chitChat.')
	print("Leave empty if you don't have any friends. ;(")
	response = input('Please leave a space between each username. ')
	if response == 'back': return False # To go back to previous layer

	names = response.split()

	''' Creating arguments for register API call '''
	argKeys = ['username', 'password', 'display_name', 'friend1', 'friend2', 'friend3', 'friend4', 'friend5']
	argValues = [username, password, display_name]
	for name in names: argValues.append(name)
	args = dict(zip(argKeys, argValues))

	''' Making the API call '''
	response = requests.post(BASE + "register", args)
	result = response.json()

	''' If username taken, try again. If not, login '''
	if 'user_added' in result:
		print('')
		return True

	else: print('Sorry, this username is taken, please try a different username.')

def login():
	username = input('Please provide a username. ')
	if username == 'back': return ['back', None] # This will allow the login/register screen to reset

	password = input('Please provide a password. ')
	if password == 'back': return ['back', None] # This will allow the login/register screen to reset

	''' Creating arguments for register API call '''
	argKeys = ['username', 'password']
	argValues = [username, password]
	args = dict(zip(argKeys, argValues))

	''' Making the API call '''
	response = requests.get(BASE + "login", args)
	result = response.json()

	''' Respond appropriately '''
	if result == {'success': 'Passwords match.'}:
		return [True, username]

	elif result == {'error': 'Passwords do not match.'}:
		print('Sorry, your password is incorrect, please try again.')

	elif result == {'error': 'This user does not exist.'}:
		print('Sorry, this user does not exist, please try another username.')

def border(message, where='topAndBottom', border='-'):
	
	length = len(message)

	if where == 'top':
		print(length * border)
		print(message)

	elif where == 'bottom':
		print(message)
		print(length * border)

	else:
		print(length * border)
		print(message)
		print(length * border)
	
def home(username):

	''' Find conversations via API call '''
	response = requests.get(BASE + "home", {'username': username})
	result = response.json()

	''' Check if there are any conversations '''
	if 'None' in result:

		toPrint = ["You don't have any conversations, why don't you start a new one?", '']
		return [None, toPrint]

	else: 

		toPrint = ["Here are your ongoing conversations.", '']

		''' Present conversations nicely '''
		chatPoss = [] # This will be useful for chat_name API call

		for convo in result:
			friend = result[convo][0]
			friendUserName = convo
			chatPoss.append([friend, friendUserName])
			message = result[convo][1][:35]
			space_size = 15-len(friend)
			spaces = " " * space_size

			toPrint.append("{}{}{}".format(friend, spaces, message))

		return [chatPoss, toPrint]

def chat_name(username, friendUserName, friendDisplayName):
	''' API call to get chat messages '''
	response = requests.get(BASE + "chat_name", {'username':username, 'friend':friendUserName})
	result = response.json()

	convo = result['conversation']

	''' Present nicely '''
	border('Here is your conversation with {}'.format(friendDisplayName))
	[print("{}>> {}".format(message[0],message[1])) for message in convo]

def chat_new(friend_name):

	''' Create new conversation using API call '''
	response = requests.post(BASE + "chat_new", {'username':username, 'friend_name': friend_name})

	''' API call to find display name for friend '''
	response = requests.get(BASE + "display_name", {'username': friend_name})
	result = response.json()

	try:
		friend_display_name = result['display_name']
		return friend_display_name

	except:
		return False


''' Pathway '''
border('Welcome to chitChat!')

while True: # Layer 1 - login/register/leave

	loggedIn = False

	print('Would you like to login, register or leave? ')
	print("(Type 'back' to return to previous stage at any time)")
	print('')
	response = input('')
	print('') # Line space

	''' Handling response '''
	if response == 'leave': break

	elif response != 'leave' and response != 'login' and response != 'register':

		print('Please choose from the following options.')
		print('')
		continue

	if response == 'register':

		''' Registering '''
		stillRegistering, registered = True, False
		while stillRegistering == True: # Keep users in loop while trying to register

			register_value = register()
			if register_value != None:
				if register_value == False: 
					stillRegistering = False # If user pressed 'back', they restart process
					print('')

				else: # User successfully registers
					stillRegistering, registered = False, True

			else: print('') # If user entered incorrect values

		if registered == False: continue # Sends users back to login/register


	''' Logging in '''
	stillLoggingIn, loggedIn = True, False
	while stillLoggingIn == True: # Keep users in loop while trying to register

		try:
			success, username = login()

			if [success, username] == ['back', None]: 
				stillLoggingIn = False
				print('')
				 
			else: # User successfully logs in
				stillLoggingIn, loggedIn = False, True

		except: pass # If user entered incorrect values

	if loggedIn == False: continue # Sends users back to login/register


	border('Welcome {}, you are now logged in!'.format(username))

	while True: # Layer 2 - Home screen (view conversations)

		chatPoss, toPrint = home(username)
		
		for line in toPrint:
			print(line)

		if chatPoss != None: # User has ongoing conversations
			
			display_names = [poss[0] for poss in chatPoss] # Display names for friends user has conversations with

		else: # There are no ongoing conversations

			display_names = []

		print('')
		print("Chat with a friend in your conversations list by typing their display name,")
		print("start a new chat by typing 'new'")
		print("check for new messages with 'refresh'")
		print("or log-out of chat server by typing 'log-out'")
		print('')

		response = input('')

		if response == 'log-out': 

			print('')
			break # Takes users back to login/registration

		elif response == 'refresh': 

			print('')
			continue

		elif response not in display_names and response != 'new':

			print('Please choose from the following options.')
			print('')
			continue

		elif response == 'new': # User wants to start a new chat

			stillFindingFriend, foundFriend = True, False
			while stillFindingFriend == True: # Keep users in loop till they have found a friend to chat with

				''' Ask for friend's name '''
				maybeFriend = input('Please give the username for a friend you would like to chat to. ')
				print('')
				if maybeFriend == 'back': stillFindingFriend = False

				else: # Actual response

					''' Check if maybeFriend is actually a friend of their's '''
					result = requests.get(BASE + "friends_info", {'username': username}).json()
					friends = result['friends_info']

					if maybeFriend in friends:
						
						''' Create new conversation and jump into it '''
						friend_name = maybeFriend
						friend_display_name = chat_new(friend_name) # Create variable response so that it can be used in chat_name
						stillFindingFriend, foundFriend = False, True

					else: # User gave an invalid username

						print('You do not have a friend by that username, please try a different username.')
						print('')

			if foundFriend == False: continue # Sends user back to home screen
			if foundFriend == True: response = friend_display_name # Create variable response so that it can be used in chat_name


		''' Jump into an ongoing conversation '''
		chatPoss, toPrint = home(username) # Check if there are any new conversations added
		display_names = [poss[0] for poss in chatPoss] # Display names for friends user has conversations with
		user_names = [poss[1] for poss in chatPoss] # Display names for friends user has conversations with

		index = 0
		for name in display_names:
			if name == response:
				break
			index += 1
		friend_user_name = user_names[index] # Inelegant way of finding friend_user_name

		chat_name(username, friend_user_name, response)
		print('')


		while True: # Layer 3 - Sending messages

			''' Check for commands '''
			message = input('>> ')
			if message == 'back': 

				print('')
				break # Send user back to home screen

			if message == 'refresh': 

				chat_name(username, friend_user_name, response)
				print('')
				continue

			''' Send message '''
			requests.post(BASE + "send", {'username':username, 'friend_name':friend_user_name, 'message':message})
			chat_name(username, friend_user_name, response)
			print('')
			
		





				


				
				


		# ''' Be able to send messages '''
		# while True: # User should be able to send as many messages as they please

		# 	''' Send message '''
		# 	message = input('>> ')
		# 	requests.post(BASE + "send", {'username':username, 'friend_name':friend_name, 'message':message})
		# 	chat_name(username, friend_name, friend_display_name)
































# # while loggedIn == False:
# # 	if response == 'login':

# # 		try:
# # 			success, username = login()
# # 			if [success, username] == ['back', None]: # For if you want to return to login/register
# # 				response = 'else'
# # 				print('')
			
# # 			else: loggedIn = True

# # 		except: pass
		
# # 	elif response == 'register':
# # 		if register() == True: 
# # 			print('Please enter your log-in details.')
			
# # 			while True:

# # 				try:
# # 					success, username = login()
# # 					loggedIn = True
# # 					break

# # 				except: pass
					
# # 	elif response == 'leave':
# # 		print('Press Ctrl+C to leave chitChat.')
# # 		break

# # 	else: response = input('Please choose one of the following: "login", "register", "leave" ')


# ''' Home screen '''
# border('Welcome {}, you are now logged in!'.format(username))
# chatPoss = home(username) # To check input against
# print('')

# ''' User input on home screen '''
# while True:

# 	print("Chat with a friend in your conversations list by typing their display name,")
# 	print("start a new chat by typing 'new'") 
# 	print("or leave chat server by typing 'leave'")
# 	print('')
# 	response = input('')

# 	if response == 'leave': break # This will change when it is in a while loop

# 	elif response == 'new': 

# 		maybeFriend = input('Please give the username for which friend you would like to chat to. ') 
# 		print('')
		
# 		''' Call API to find which friends the user has '''
# 		result = requests.get(BASE + "friends_info", {'username': username}).json()

# 		''' Check if maybeFriend is actually a friend of their's '''
# 		friends = result['friends_info']
# 		if maybeFriend in friends:
			
# 			''' Create new conversation and jump into it '''
# 			friend_name = maybeFriend
# 			friend_display_name = chat_new(friend_name)
# 			chat_name(username, friend_name, friend_display_name)

# 			''' Be able to send messages '''
# 			while True: # User should be able to send as many messages as they please

# 				''' Send message '''
# 				message = input('>>')
# 				requests.post(BASE + "send", {'username':username, 'friend_name':friend_name, 'message':message})
# 				chat_name(username, friend_name, friend_display_name)

# 		else: 
# 			print('You do not have a friend with that username, please choose someone else.')
# 			print('')

# 	else: # User may have wanted to view a chat

# 		''' Work out whether user inputted a friend's name '''
# 		friendHere = False

# 		try:
# 			for poss in chatPoss:
# 				friendDisplayName = poss[0]
# 				friendUserName = poss[1]

# 				if response == friendDisplayName:
# 					friendHere = True
# 					break

# 		except:
# 			pass

# 		if friendHere == True:
# 			''' API call to get chat messages '''
# 			chat_name(username, friendUserName, friendDisplayName)
			
# 			''' Be able to send messages '''
# 			while True: # User should be able to send as many messages as they please

# 				''' Send message '''
# 				message = input('>> ')
# 				requests.post(BASE + "send", {'username':username, 'friend_name':friend_name, 'message':message})
# 				chat_name(username, friend_name, friend_display_name)

# 		else:
# 			print('Please try one of the following options')
# 			print('') # Line break




# # Scenarios where you need to be able to go back:
# # Everywhere where there is user input






















# userPathway

# get login_register - brings up log-in / register page 

# while True: # You log-out when you press leave

# 	user input - login, register or press Ctrl+C to exit the server
# 	if login:
# 		user input - please give username and password
# 		get login - lets you in, passing in username and password

# 	elif register:
# 		user input - provide username, password and repeated password
# 		post register - creates new user in database
# 		user input - please give username and password
# 		get login - lets you in, passing in username and password

# 	while True:

# 		get home - brings up conversations
# 		user input - your options are: chat [name], chat [new], leave
# 		if chat [name]:
			
# 			(chat_name function)
# 			while True:

# 				get chat_name - brings up conversation, pass in name
# 				user input - enter send to send a message, refresh to refresh page

# 				if send:
# 					user input - >> 
# 					post send - adds message to conversation, passing in message

# 				elif refresh:
# 					pass

# 				elif home:
# 					break

# 		elif chat [new]:
# 			get friends_names - brings up list of friends names
# 			user input - please give name
# 			post chat_new - adds conversation to conversations list
# 			run chat_name function again

# 		elif leave:
# 			break



from flask import Flask
from flask_restful import Api, Resource, reqparse
import sqlite3
import bcrypt
# import createDB

''' SQL Functions '''
def newUser(username, password, display_name, friend1, friend2, friend3, friend4, friend5):
	
	conn = sqlite3.connect('users.db') # Setting up the database (and where)
	c = conn.cursor() # This allows us to run commands to database

	password = password.encode('utf-8')
	password = bcrypt.hashpw(password, bcrypt.gensalt())

	with conn:
		c.execute("INSERT INTO users VALUES(?,?,?,?,?,?,?,?)", (username, password, 
																display_name, friend1, 
																friend2, friend3, 
																friend4, friend5))
	
	conn.close()

def findUser(username):
	
	conn = sqlite3.connect('users.db') # Setting up the database (and where)
	c = conn.cursor() # This allows us to run commands to database

	c.execute("SELECT * FROM users WHERE username=?", (username,))
	results = c.fetchall()
	return results

	conn.close()

def findAll():

	conn = sqlite3.connect('users.db') # Setting up the database (and where)
	c = conn.cursor() # This allows us to run commands to database

	c.execute("SELECT * FROM users")
	results = c.fetchall()
	return results

	conn.close()

def delUser(username):

	conn = sqlite3.connect('users.db') # Setting up the database (and where)
	c = conn.cursor() # This allows us to run commands to database

	with conn:
		c.execute("DELETE FROM users WHERE username=?", (username,))

	conn.close()


''' Initialising the server with API '''
app = Flask(__name__) # 0.0.0.0 to hear from any computer
api = Api(app)


''' Conversations holder '''
convos = [['testuser1', 'testuser2', [['testuser1', 'message1'], ['testuser2', 'message2'],['testuser1',"here's another message"]]],
		['testuser1', 'testuser3', [['testuser3', 'message1'], ['testuser1', 'message2'], ['testuser1',"how are you today?"]]],
		['testuser1', 'thutch97', [['testuser1', 'What is my purpose?'], ['thutch97', 'You pass butter'], ['testuser1',"Oh my God"]]],
		['testuser2', 'testuser3', [['testuser2', 'message1'], ['testuser3', 'message2'],['testuser2',"here's another message"]]]]


''' API Resources '''
class Login(Resource):
	''' Login resource '''

	def get(self):
		
		''' Define arguments for get request '''
		login_get_args = reqparse.RequestParser()
		login_get_args.add_argument('username', type=str, help='Username required', required=True)
		login_get_args.add_argument('password', type=str, help='Password required', required=True)
		args = login_get_args.parse_args()

		''' Get user information '''
		givenUsername = args['username']
		givenPassword = args['password'].encode('utf-8')
		userInfo = findUser(givenUsername)
		
		if userInfo != []:
			password = userInfo[0][1]
		
		else: return {'error': 'This user does not exist.'}

		''' Check password '''
		if bcrypt.checkpw(givenPassword, password):
			return {'success': 'Passwords match.'}
		else:
			return {'error': 'Passwords do not match.'}
api.add_resource(Login, "/login")


class Register(Resource):
	''' Register resource '''

	def post(self):
		
		''' Define arguments for post request '''
		register_post_args = reqparse.RequestParser()
		register_post_args.add_argument('username', type=str, help='Username required', required=True)
		register_post_args.add_argument('password', type=str, help='Password required', required=True)
		register_post_args.add_argument('display_name', type=str, help='Display name required', required=True)
		register_post_args.add_argument('friend1', type=str, help='friend1 required')
		register_post_args.add_argument('friend2', type=str, help='friend2 required')
		register_post_args.add_argument('friend3', type=str, help='friend3 required')
		register_post_args.add_argument('friend4', type=str, help='friend4 required')
		register_post_args.add_argument('friend5', type=str, help='friend5 required')

		args = register_post_args.parse_args()

		''' Check if username already taken '''
		username = args['username']
		if findUser(username) != []: return {'message': 'Username already taken'}

		else:

			''' Create new user in the database '''
			password = args['password']
			display_name, friend1 = args['display_name'], args['friend1']
			friend2, friend3 = args['friend2'], args['friend3']
			friend4, friend5 = args['friend4'], args['friend5']
			
			newUser(username, password, display_name, 
					friend1, friend2, friend3, friend4, friend5)

			''' To check new user has been added '''
			if findUser(username) != []: added = True
			else: added = False 

			return {'user_added': added}
api.add_resource(Register, "/register")


class Home(Resource):
	''' Home resource '''

	def get(self):

		''' Define arguments for get request '''
		login_get_args = reqparse.RequestParser()
		login_get_args.add_argument('username', type=str, help='Username required', required=True)
		args = login_get_args.parse_args()

		username = args['username']

		''' Display conversations '''
		myConversations = []

		for conversation in convos:

			if username in conversation:
				myConversations.append(conversation)

		if myConversations == []: 
			return {'None': 'No conversations'}

		else:

			toReturn = {}

			for conversation in myConversations:

				for item in conversation:
					if item != username: 
						friend = item
						break

				friendDisplayName = findUser(friend)[0][2]
				messages = conversation[-1]

				''' Different result if conversation is empty '''
				try:
					lastMessage = messages[-1]
					lastMessageContent = lastMessage[-1]
					toReturn[friend] = [friendDisplayName,lastMessageContent]

				except:
					toReturn[friend] = [friendDisplayName,' ']

			return toReturn
api.add_resource(Home, "/home")


class Chat_name(Resource):
	''' Chat_name resource '''

	def get(self):
		''' Give conversation when given username '''

		''' Define arguments for get request '''
		login_get_args = reqparse.RequestParser()
		login_get_args.add_argument('username', type=str, help='Username required', required=True)
		login_get_args.add_argument('friend', type=str, help='Friend required', required=True)
		args = login_get_args.parse_args()

		username = args['username']
		friend = args['friend']

		''' Get conversation from convos list '''
		myConversations = []

		for conversation in convos:

			if username in conversation:
				myConversations.append(conversation)

		for item in myConversations:
			if friend in item: 
				yesConvo = item # Indicating this is the correct conversation
				break

		''' Return conversation '''
		try:
			messages = yesConvo[-1]
			return {'conversation': messages}

		except:
			return {'None': 'This conversation does not exist.'}
api.add_resource(Chat_name, "/chat_name")


class Friends_info(Resource):

	def get(self):
		''' Give conversation when given username '''

		''' Define arguments for get request '''
		login_get_args = reqparse.RequestParser()
		login_get_args.add_argument('username', type=str, help='Username required', required=True)
		args = login_get_args.parse_args()

		username = args['username']

		''' Get friends info for this username '''
		try:
			friends_info = findUser(username)[0][3:]
			return {'friends_info':friends_info}

		except:
			return {'None': 'There is no user by this username.'}
api.add_resource(Friends_info, "/friends_info")


class Chat_new(Resource):
	''' Chat_new resource '''

	def post(self):

		''' Define arguments for get request '''
		login_get_args = reqparse.RequestParser()
		login_get_args.add_argument('username', type=str, help='Username required', required=True)
		login_get_args.add_argument('friend_name', type=str, help='Friend name required', required=True)
		args = login_get_args.parse_args()

		username = args['username']
		friend_name = args['friend_name']

		''' Find username's conversations '''
		myConversations = []

		for conversation in convos:

			if username in conversation:
				myConversations.append(conversation)

		''' Check if conversation with friend already exists '''
		for convo in myConversations:
			if friend_name in convo:
				return {'Already exists': 'This conversation already exists.'}

		''' Create new conversation if this conversation doesn't already exist '''
		convos.append([username, friend_name,[]])
		return {'New': 'New conversation set up.'}
api.add_resource(Chat_new, "/chat_new")


class Display_name(Resource):

	def get(self):
		''' Get display name if given username '''

		''' Define arguments for get request '''
		login_get_args = reqparse.RequestParser()
		login_get_args.add_argument('username', type=str, help='Username required', required=True)
		args = login_get_args.parse_args()

		username = args['username']

		''' Get display name '''
		try:
			display_name = findUser(username)[0][2]
			return {'display_name':display_name}

		except:
			return {'None': 'There is no user by this username.'}
api.add_resource(Display_name, "/display_name")


class Send(Resource):
	''' Send resource '''
	
	def post(self):

		''' Define arguments for get request '''
		login_get_args = reqparse.RequestParser()
		login_get_args.add_argument('username', type=str, help='Username required', required=True)
		login_get_args.add_argument('friend_name', type=str, help='Friend name required', required=True)
		login_get_args.add_argument('message', type=str, help='Message required', required=True)
		args = login_get_args.parse_args()

		username = args['username']
		friend_name = args['friend_name']
		message = args['message']

		''' Find relevant conversation '''
		index, found = 0, False
		for convo in convos:

			if username in convo and friend_name in convo:
				
				yesConvoIndex, found = index, True
				break
			index += 1

		if found == True: # Ensuring the correct message has been found

			''' Add message to conversation '''
			yesConvo = convos[yesConvoIndex] # Find correct conversation
			messages = yesConvo[-1]
			messages.append([username,message])

			return {'Sent': yesConvo}

		else:
			return {'None': 'Message not sent'}
api.add_resource(Send, "/send")


if __name__ == '__main__':
	app.run(debug=True)


import sqlite3
import bcrypt

''' Instantiate database '''
conn = sqlite3.connect('users.db') # Setting up the database (and where)
c = conn.cursor() # This allows us to run commands to database

''' Create table '''
c.execute("""CREATE TABLE users (
 			username text,
 			password text,
 			display_name text,
 			friend1 text,
 			friend2 text,
 			friend3 text,
 			friend4 text,
 			friend5 text
 			)""")

''' Create test entries '''
thutch97_password = 'myPass'.encode('utf-8')
thutch97_password = bcrypt.hashpw(thutch97_password, bcrypt.gensalt())

rabbit_password = '123'.encode('utf-8')
rabbit_password = bcrypt.hashpw(rabbit_password, bcrypt.gensalt())

testuser1_password = 'testusr1'.encode('utf-8')
testuser1_password = bcrypt.hashpw(testuser1_password, bcrypt.gensalt())

testuser2_password = 'testusr2'.encode('utf-8')
testuser2_password = bcrypt.hashpw(testuser2_password, bcrypt.gensalt())

testuser3_password = 'testusr3'.encode('utf-8')
testuser3_password = bcrypt.hashpw(testuser3_password, bcrypt.gensalt())

with conn:
	c.execute("INSERT INTO users VALUES('thutch97',?,'Tom','rabbit','testuser1','testuser2','testuser3', 'None')", (thutch97_password,))
	c.execute("INSERT INTO users VALUES('rabbit',?,'Peter','thutch97','None','None','None','None')", (rabbit_password,))
	c.execute("INSERT INTO users VALUES('testuser1',?,'testo','thutch97','testuser2','testuser3','None', 'None')", (testuser1_password,))
	c.execute("INSERT INTO users VALUES('testuser2',?,'testinho','thutch97','testuser1','testuser3','None', 'None')", (testuser2_password,))
	c.execute("INSERT INTO users VALUES('testuser3',?,'testy','thutch97','testuser1','testuser2','None', 'None')", (testuser3_password,))

conn.close()


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
      name='chitChat-pkg-thutch97',
      version='0.0.4',
      author='Tom Hutchinson',
      author_email='thomas.3.hutchinson@bt.com',
      description='A python chat server package.',
      long_description=long_description,
      long_description_content_type="text/markdown",
      install_requires = [
            'Flask==1.1.2',
		'Flask-RESTful==0.3.8',
		'requests==2.24.0',
		'bcrypt==3.2.0'
      ],
      packages=setuptools.find_packages(),
      classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
      ],
      python_requires='>=3.6',
      package_data={
            'chitChat-pkg-thutch97':['users.db']
      }
     )